
function auto_align_images_func,pq

   ; This routine calculates the cross correlation between the reference
   ; image and the transformed image after applying the pq transformation.
   ; It is used in the minimization routine to find the best transformation
   ; and should not be called directly.

   common auto_align_images_private1,rimage,timage, $
                                    npq,rnx,rny,nxt,nyt, $
                                    restricted,quiet,xmask,ymask,rmask

   if restricted then begin
      ;pq = [exscl,eyscl,exshft,eyshft,erot]
      tr = rss2pq(nxt,nyt,xshift=pq[2],yshift=pq[3],rot12=pq[4], $
                  xscale=pq[0],yscale=pq[1],/center,p=p,q=q)
   endif else begin
      p = pq[0L:npq-1L]
      q = pq[npq:*]
   endelse
   missing = long(min(timage)-1000.)
   itest = poly_2d(timage,p,q,1,rnx,rny,missing=missing)

   good = where(itest GT missing,ngood)
   if ngood GT 0 then begin

      ; Make the arrays as small as possible to increase speed
      x = good MOD rnx
      y = good / rnx
      minx = min(x) & maxx = max(x)
      miny = min(y) & maxy = max(y)
      snx = maxx-minx+1L
      sny = maxy-miny+1L
      ; Cut off zero areas to save computation time
      srimage = rimage[minx:maxx,miny:maxy]
      sitest = itest[minx:maxx,miny:maxy]
      nxy = n_elements(srimage)

      mask = fltarr(snx,sny)  ; Mask out all the "bad" points
      mask[*] = 1.0
      bad = where(sitest LE missing,nbad)
      if nbad GT 0 then mask[bad] = 0.0 ; Zero out missing areas
      if xmask[0] GE 0 then begin  ; Apply transformed mask
         p = reform(p,2,2)
         q = reform(q,2,2)
         tforward = fltarr(4,2)
         tforward[*,0] = p
         tforward[*,1] = q
         treverse=caltrans(ppinvert(pq2pp(tforward))) ; Invert p&q
         p = reform(treverse[*,0],2,2)  ; inverted p & q
         q = reform(treverse[*,1],2,2)
         ; Get the transformed coordinates
         txmask = p(0,0) + p(0,1)*xmask + p(1,0)*ymask + p(1,1)*xmask*ymask - minx
         tymask = q(0,0) + q(0,1)*xmask + q(1,0)*ymask + q(1,1)*xmask*ymask - miny
         mgood = where(txmask GE 0 AND txmask LE (snx-1) AND $
                       tymask GE 0 AND tymask LE (sny-1),nmgood)
         if nmgood GT 0 then begin
            mask[txmask[mgood],tymask[mgood]] = 0.0
            if NOT keyword_set(quiet) then $
               srimage[txmask[mgood],tymask[mgood]] = 0.0
         endif
      endif
      if rmask[0] GE 0 then begin  ; Apply the reference mask
         xrmask = (rmask MOD rnx) - minx
         yrmask = (rmask / rnx) - miny
         mgood = where(xrmask GE 0 AND xrmask LE (snx-1) AND $
                       yrmask GE 0 AND yrmask LE (sny-1),nmgood)
         if nmgood GT 0 then begin
            mask[xrmask[mgood],yrmask[mgood]] = 0.0
            if NOT keyword_set(quiet) then $
               srimage[xrmask[mgood],yrmask[mgood]] = 0.0
         endif
      endif

      ; Compute the cross correllation at the good points
      cgood = where(mask NE 0.0,ncgood)
      if ncgood EQ 0 then crosscorr = 0. $
      else begin
         ravg = total(srimage[cgood])/ncgood
         tavg = total(sitest[cgood]) /ncgood
         crosscorr = ((srimage-ravg))*((sitest-tavg))*mask
         crosscorr = total(crosscorr)/ncgood
      endelse

      ; Optionally show the progress
      if NOT keyword_set(quiet) then begin
         if !d.x_size LT snx OR !d.y_size LT sny then begin
            if !d.window GE 0 then wdelete,!d.window
            wdef,ws,snx,sny
         endif
         erase
         tvscl,srimage
         ocontour2,sitest,level=max(sitest)*[0.1,0.25,0.50,0.75]
         xyouts,/normal,0.0,0.0,string(crosscorr[0,0]),align=0.0
      endif

      return,-crosscorr[0,0]  ; negative since powel minimizes, 
                              ; but we want to maximize the cross corr
   endif else return,0.0  ; No good points so must be WAY off.

end

;+

function auto_align_images,timagein,rimagein,pin,qin,pout,qout, $
                           unrestricted=unrestrictedin,quiet=quietin, $
                           missing=missval,itmax=itmaxin,double=double, $
                           tmask=tmask,rmask=rmaskin

;NAME:
;     AUTO_ALIGN_IMAGES
;PURPOSE:
;     Align arbitrary images using a cross correlation.  This routine is
;     particularly useful for images from different instruments in which
;     a very general transformation between the two images is required.  
;     Don't use this routine if you only need to shift the two images on
;     top of one another.  It'll work, but it will be much slower than
;     you need.
;CATEGORY:
;CALLING SEQUENCE:
;     inew=auto_align_images(transformed_image,reference_image, $
;                            pin,qin,pout,qout)
;INPUTS:
;     transformed_image = image to be transformed onto the reference image.
;     refererence_image = reference image
;     pin,qin = Initial guess for the p and q passed to poly_2d.  You must get
;               this pretty close!! e.g. you can use the mouse to click on 
;               corresponding points like this:
;           pp = setpts_roi(transformed_image,reference_image)
;           tt = caltrans(pp)
;           pin = tt[*,0]
;           qin = tt[*,1]
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     missing = Value to use where the transformed image is out of the 
;               field of view of the initial image to be transformed.
;     /unrestricted = Do not restrict the transformation to rotation, shift, 
;                     and scale, i.e. allow a general very general warping.
;                     Chances are you do not want to set this keyword in most 
;                     circumstances.  You might use it if the images have
;                     atmospheric blurring or something like that.
;     tmask = index vector into the transformed image giving pixels which are
;             to be ignored in the transformed image when computing the cross
;             correlation.  Use to wipe out off-limb ponts, blemishes, etc.
;     rmask = index vector into the reference image giving pixels which are
;             to be ignored in the reference image when computing the cross
;             correlation.  Use to wipe out off-limb points, blemishes, etc.
;     itmax = maximum number of iterations (default = 200).
;     /double = use double precision in the minimization.
;     /quiet = work quietly
;OUTPUTS:
;     inew = new transformed image overlaid in reference image.
;     pout,qout = Output transformation.  
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     The initial guess (pin,qin) must be close.  This routine only refines
;     that guess.
;
;     *Very slow*
;PROCEDURE:
;
;     The initial warping is tweaked until a maximum in the cross correlation
;     is found using the powell minimization procedure.
;
;     The transformed image overlaid on the reference image is:
;
;     new_image = poly_2d(transformed_image,pout,qout,1,rnx,rny,missing=0.0)
;     tvscl,reference_image
;     ocontour2,new_image
;
;        where rnx,rny is the size of the reference image
;
;MODIFICATION HISTORY:
;     T. Metcalf 2001-09-25
;     T. Metcalf 2001-09-26 Improved the way the display windows work.  If the
;                           window is too small, it will appear the that 
;                           alignment is wrong even though it is right.  So,
;                           make sure that the windows are always large
;                           enough.
;     T. Metcalf 2001-09-28 Added tmask and rmask keywords.
;-

common auto_align_images_private1,rimage,timage, $
                                 npq,rnx,rny,nxt,nyt, $
                                 restricted,quiet,xmask,ymask,rmask

if keyword_set(itmaxin) then itmax=itmaxin else itmax=200
if keyword_set(unrestrictedin) then restricted=0 else restricted=1
if keyword_set(quietin) then quiet=1 else quiet=0
if keyword_set(rmaskin) then rmask = rmaskin else rmask = -1

rnx = n_elements(rimagein[*,0])
rny = n_elements(rimagein[0,*])
nxt = n_elements(timagein[*,0])
nyt = n_elements(timagein[0,*])

; In case xmask & ymask are left in the common block from a previous call
xmask = -1 & ymask = -1

if keyword_set(tmask) then begin  ;set up the transformed image mask
   if tmask[0] GE 0 then begin
      xmask = tmask MOD nxt
      ymask = tmask / nxt
   endif
endif

rimage = rimagein
timage = timagein

; The restricted transformation includes only rotation, scale, and shift.
; The unrestricted transformation also includes warping.
if restricted then begin
   if NOT quiet then begin
      print
      print,'Initial guess:'
   endif
   ; Get the initial guess for rotation, shift, and scale.
   ; enrss gives the "leftover" warping in the initial guess
   ; that is neglected.
   pq2rss,pin,qin,erot,exscl,eyscl,exshft,eyshft,enrss,nxt,nyt, $
          /center,quiet=quiet
   pq = [exscl,eyscl,exshft,eyshft,erot]
   npq = -1
endif else begin
   npq = n_elements(pin)
   if n_elements(qin) NE npq then $
      message,'p and q must have the same dimension'
   pq = [pin,qin]
endelse
npq2 = n_elements(pq)

xi = fltarr(npq2,npq2)
for i=0L,npq2-1L do xi[i,i]=1.0   ; unit vectors for Powell minimization

; Now do the minimization to find the best alignment

powell,pq,xi,1.e-4,mincorr,'auto_align_images_func', $
       iter=niter,itmax=200,double=double

if niter GE itmax then message,/info,'WARNING: Failed to fully converge'

; Get the final transformation for output
if restricted then begin
   ;pq = [exscl,eyscl,exshft,eyshft,erot]
   tr = rss2pq(nxt,nyt,xshift=pq[2],yshift=pq[3],rot12=pq[4], $
               xscale=pq[0],yscale=pq[1],/center,p=p,q=q)
endif else begin
   p = pq[0L:npq-1L]
   q = pq[npq:*]
endelse

itest = poly_2d(timage,p,q,1,rnx,rny,missing=missval) ; The transformed image

if NOT quiet then begin  ; Optionally display the alignment
   if !d.x_size LT rnx OR !d.y_size LT rny then wdef,ws,rnx,rny
   tvscl,rimage
   ocontour2,itest,level=max(itest)*[0.1,0.25,0.50,0.75,0.90]
endif

; Optionally display rotation, shift, and scale
if restricted and NOT quiet then begin 
   print
   print,'Final parameters:'
   pq2rss,p,q,erot,exscl,eyscl,exshft,eyshft,enrss,nxt,nyt,/center
end

pout = p
qout = q
return,itest

end
